@@ -1,9 +1,9 @@
 import logging
-import os
 from abc import ABC, abstractmethod
 from pathlib import Path
-from typing import Any, Callable, Dict, List, Optional, Type, Union, cast
+from typing import Any, Callable, Dict, List, Optional, Tuple, Type, Union, cast
 
+import boto3
 from encode_utils.connection import Connection
 
 from accession.accession_steps import (
@@ -23,7 +23,7 @@ from accession.encode_models import (
     EncodeStepRun,
 )
 from accession.file import GSFile
-from accession.helpers import flatten, string_to_number
+from accession.helpers import LruCache, flatten, string_to_number
 from accession.logger_factory import logger_factory
 from accession.preflight import MatchingMd5Record, PreflightHelper
 
@@ -49,10 +49,13 @@ class Accession(ABC):
         self.conn = connection
         self.common_metadata = common_metadata
         self.new_files: List[EncodeFile] = []
+        self.upload_queue: List[Tuple[EncodeFile, GSFile]] = []
         self.new_qcs: List[Dict[str, Any]] = []
         self.raw_qcs: List[EncodeQualityMetric] = []
         self.log_file_path = log_file_path
         self.no_log_file: bool = no_log_file
+        # keys are hex md5sums, values are lists of portal objects
+        self.search_cache: LruCache[str, List[Dict[str, Any]]] = LruCache()
         self._logger: Optional[logging.Logger] = None
         self._experiment: Optional[EncodeExperiment] = None
         self._preflight_helper: Optional[PreflightHelper] = None
@@ -108,7 +111,7 @@ class Accession(ABC):
     def experiment(self) -> EncodeExperiment:
         if self._experiment is None:
             encode_file = self.get_encode_file_matching_md5_of_blob(
-                self.analysis.raw_fastqs[0].filename
+                self.analysis.raw_fastqs[0]
             )
             if encode_file is None:
                 raise ValueError("Could not find raw fastqs on the portal")
@@ -120,11 +123,31 @@ class Accession(ABC):
         self, file: GSFile
     ) -> Optional[List[EncodeFile]]:
         """
-        Retrieves all files from the portal with an md5sum matching the blob's md5
+        Retrieves all files from the portal with an md5sum matching the blob's md5. Will
+        always attempt to use cached results. We need to search with frame=embedded so
+        that the portal will return the full file objects, otherwise they will return
+        with an arbitrary frame that does not include even the md5sum.
         """
-        md5sum = self.backend.md5sum(file)
-        search_param = [("md5sum", md5sum), ("type", "File")]
-        encode_files = self.conn.search(search_param)
+        file_md5sum = file.md5sum
+        search_param = [
+            ("md5sum", file_md5sum),
+            ("type", "File"),
+            ("frame", "embedded"),
+        ]
+        cache_result = self.search_cache.get(file_md5sum)
+        # Handle cache miss
+        if cache_result is None:
+            self.logger.debug(
+                "Could not retrive search result from cache for md5sum %s will search portal",
+                file_md5sum,
+            )
+            encode_files = self.conn.search(search_param)
+            self.search_cache.insert(file_md5sum, encode_files)
+        else:
+            self.logger.debug(
+                "Will use cached search result for file with md5sum %s", file_md5sum
+            )
+            encode_files = cache_result
         if not encode_files:
             return None
         modeled_encode_files = [EncodeFile(file_props) for file_props in encode_files]
@@ -136,7 +159,7 @@ class Accession(ABC):
         """Finds an ENCODE File object whose md5sum matches md5 of a blob in URI in backend.
 
         Args:
-            file (str): String representing an URI to an object in the backend.
+            file (GSFile): A GSFile representing an object on the backend.
 
         Returns:
             EncodeFile: an instance of EncodeFile, a document-object mapping
@@ -151,8 +174,7 @@ class Accession(ABC):
                 self.logger.warning(
                     "get_encode_file_matching_md5_of_blob found more than 1 files matching the md5 of the blob."
                 )
-            encode_file = self.conn.get(filtered_encode_files[0].at_id)
-            return EncodeFile(encode_file)
+            return filtered_encode_files[0]
         else:
             return None
 
@@ -164,7 +186,7 @@ class Accession(ABC):
         preflight helper would be required to know about the method
         `get_all_encode_files_matching_md5_of_blob`.
         """
-        matching = self.get_all_encode_files_matching_md5_of_blob(gs_file.filename)
+        matching = self.get_all_encode_files_matching_md5_of_blob(gs_file)
         if matching is None:
             return None
         record = self.preflight_helper.make_file_matching_md5_record(
@@ -174,59 +196,80 @@ class Accession(ABC):
 
     def raw_files_accessioned(self):
         for file in self.analysis.raw_fastqs:
-            if not self.get_encode_file_matching_md5_of_blob(file.filename):
+            if not self.get_encode_file_matching_md5_of_blob(file):
                 return False
         return True
 
     def accession_file(
         self, encode_file: Dict[str, Any], gs_file: GSFile
     ) -> EncodeFile:
-        file_exists = self.get_encode_file_matching_md5_of_blob(gs_file.filename)
-        submitted_file_path = {"submitted_file_name": gs_file.filename}
+        """
+        First POSTs the file metadata and subsequently queues upload of the actual data.
+        The file is queued for upload if there are no 409 conflicts for the posted file
+        metadata. In addition, if there is a conflict and the file has a status of
+        "upload failed", then reupload will be queued. If there is a 409 conflict and
+        the file status is uploading, then we assume the file is currently being
+        uploaded and upload will not be queued.
+        """
+        file_exists = self.get_encode_file_matching_md5_of_blob(gs_file)
         if file_exists:
             self.logger.warning(
                 "Attempting to post duplicate file of %s with md5sum %s",
                 file_exists.get("accession"),
                 encode_file.get("md5sum"),
             )
-        local_file = self.backend.download(gs_file.filename)[0]
-        encode_file["submitted_file_name"] = local_file
-        encode_posted_file = self.conn.post(encode_file)
-        os.remove(local_file)
-        encode_posted_file = self.patch_file(encode_posted_file, submitted_file_path)
+        encode_posted_file, status_code = self.conn.post(
+            encode_file, upload_file=False, return_original_status_code=True
+        )
         modeled_encode_file = EncodeFile(encode_posted_file)
+        if modeled_encode_file.status == "upload failed" or (
+            modeled_encode_file.status == "uploading" and status_code != 409
+        ):
+            self.upload_queue.append((modeled_encode_file, gs_file))
+        else:
+            self.logger.info(
+                "Encode file %s is already uploaded, will not reupload",
+                modeled_encode_file.at_id,
+            )
         self.new_files.append(modeled_encode_file)
         return modeled_encode_file
 
-    def patch_file(
-        self, encode_file: Dict[str, Any], new_properties: Dict[str, Any]
-    ) -> Dict[str, Any]:
-        new_properties[self.conn.ENCID_KEY] = encode_file.get("accession")
-        return self.conn.patch(new_properties, extend_array_values=False)
-
-    def log_if_exists(self, payload):
-        """
-        If an object with given aliases already exists, as determined by an additional GET request,
-        then log a warning before attempting to POST the payload. Truthiness of the dict returned by
-        encode_utils.connection.Connection.get() is sufficient to check the object's existence on
-        the portal, since it returns an empty dict when no matching record is found.
+    def upload_file(self, encode_file: EncodeFile, gs_file: GSFile) -> None:
         """
+        At a high level, uploads the file from GCS to S3 by streaming bytes. As the s3
+        client reads chunks they are lazily fetched from GCS.
 
-        aliases = payload.get("aliases")
-        profile_key = payload[self.conn.PROFILE_KEY]
-        if aliases:
-            if self.conn.get(aliases, database=True):
-                self.logger.error(
-                    "%s with aliases %s already exists, will not post it",
-                    profile_key.capitalize().replace("_", " "),
-                    aliases,
-                )
+        In more details, obtains STS credentials to upload to the portal file specified
+        by `encode_file`, creates a s3 client, and uploads the file corresponding to
+        `gs_file` (potentially as multipart). For this to work, the blob acquired by
+        `self.backend.blob_from_filename` must return an object that has a file-like
+        `read` method. For more details see the `boto3` docs:
+        https://boto3.amazonaws.com/v1/documentation/api/latest/reference/services/s3.html#S3.Client.upload_fileobj
+
+        Extensive testing revealed that for the boto3 default transfer config performed
+        satisfactorily, see PIP-745
+        """
+        credentials = self.conn.regenerate_aws_upload_creds(encode_file.accession)
+        s3 = boto3.client(
+            "s3",
+            aws_access_key_id=credentials["access_key"],
+            aws_secret_access_key=credentials["secret_key"],
+            aws_session_token=credentials["session_token"],
+        )
+        s3_uri = credentials["upload_url"]
+        path_parts = s3_uri.replace("s3://", "").split("/")
+        bucket = path_parts.pop(0)
+        key = "/".join(path_parts)
+        filename = gs_file.filename
+        gcs_blob = self.backend.blob_from_filename(filename)
+        self.logger.info("Uploading file %s to %s", filename, s3_uri)
+        s3.upload_fileobj(gcs_blob, bucket, key)
+        self.logger.info("Finished uploading file %s", filename)
 
     def get_or_make_step_run(self, accession_step: AccessionStep) -> EncodeStepRun:
         """
-        encode_utils.connection.Connection.post() does not fail on alias conflict, and does not
-        expose the response status code, so we need to check for the existence of the object first
-        before attempting to POST it with Accession.log_if_exists().
+        encode_utils.connection.Connection.post() does not fail on alias conflict, here
+        we log if there was a 409 conflict.
         """
         docker_tag = self.analysis.get_tasks(accession_step.wdl_task_name)[
             0
@@ -240,8 +283,12 @@ class Accession(ABC):
             )
         ]
         payload = accession_step.get_portal_step_run(aliases)
-        self.log_if_exists(payload)
-        posted = self.conn.post(payload)
+        posted, status_code = self.conn.post(payload, return_original_status_code=True)
+        if status_code == 409:
+            self.logger.warning(
+                "Analysis step run with aliases %s already exists, will not post it",
+                aliases,
+            )
         return EncodeStepRun(posted)
 
     def find_portal_property_from_filekey(
@@ -253,7 +300,7 @@ class Accession(ABC):
         files = self.analysis.get_files(filekey=filekey)
         msg = "Could not find any file with key {} in metadata".format(filekey)
         if files:
-            annotation = self.get_encode_file_matching_md5_of_blob(files[0].filename)
+            annotation = self.get_encode_file_matching_md5_of_blob(files[0])
             if annotation is None:
                 raise KeyError(msg)
             return annotation.get(portal_property, "")
@@ -285,7 +332,7 @@ class Accession(ABC):
             raise
         encode_files = []
         for gs_file in derived_from_files:
-            encode_file = self.get_encode_file_matching_md5_of_blob(gs_file.filename)
+            encode_file = self.get_encode_file_matching_md5_of_blob(gs_file)
             if encode_file is not None:
                 encode_files.append(encode_file)
         accessioned_files = encode_files + self.new_files
@@ -348,6 +395,7 @@ class Accession(ABC):
             file_size=file.size,
             file_md5sum=file.md5sum,
             step_run_id=step_run.at_id,
+            submitted_file_name=file.filename,
             genome_annotation=self.genome_annotation,
             extras=extras,
         )
@@ -477,10 +525,6 @@ class Accession(ABC):
 
                     for qc in file_params.quality_metrics:
                         qc_method = getattr(self, type(self).QC_MAP[qc])  # type: ignore
-                        updated_properties = self.conn.get(
-                            encode_file.at_id, database=True
-                        )
-                        encode_file.portal_file = updated_properties
                         qc_method(encode_file, wdl_file)
                     accessioned_files.append(encode_file)
         if dry_run:
@@ -507,6 +551,8 @@ class Accession(ABC):
                 self.accession_step(step)
             self.post_qcs()
             self.patch_experiment_analyses()
+            for encode_file, gs_file in self.upload_queue:
+                self.upload_file(encode_file, gs_file)
 
 
 class AccessionGenericRna(Accession):
@@ -915,7 +961,7 @@ class AccessionChip(Accession):
             files = self.analysis.get_files(filekey)
             if not files:
                 raise ValueError(f"Could not find any files matching filekey {filekey}")
-            portal_index = self.get_encode_file_matching_md5_of_blob(files[0].filename)
+            portal_index = self.get_encode_file_matching_md5_of_blob(files[0])
             if portal_index is None:
                 raise ValueError("Could not find portal index")
             portal_assembly = portal_index.get(EncodeFile.ASSEMBLY)
@@ -1018,6 +1064,22 @@ class AccessionChip(Accession):
             ) from e
         return {"mapped_read_length": mapped_read_length}
 
+    def add_mapped_run_type(self, gs_file: GSFile) -> Dict[str, str]:
+        """
+        Obtains the value of `mapped_run_type` to post for bam files from the read
+        length log in the ancestor align task in the ChIP-seq pipeline, useful for
+        detecting PE data that was mapped as SE on the portal.
+        """
+        replicate = self.get_chip_pipeline_replicate(gs_file)
+        qc = self.backend.read_json(self.analysis.get_files("qc_json")[0])
+        is_paired_end = qc["general"]["seq_endedness"][replicate]["paired_end"]
+        if not isinstance(is_paired_end, bool):
+            raise TypeError(
+                f"Expected boolean for ChIP QC value general.seq_endedness.{replicate}.paired_end, found {is_paired_end}"
+            )
+        mapped_run_type = "paired-ended" if is_paired_end else "single-ended"
+        return {"mapped_run_type": mapped_run_type}
+
     def maybe_add_cropped_read_length(self, gs_file: GSFile) -> Dict[str, int]:
         """
         Obtains the value of mapped_read_length to post for bam files from the
@@ -1034,6 +1096,26 @@ class AccessionChip(Accession):
             return {}
         return {"cropped_read_length": crop_length}
 
+    def maybe_add_cropped_read_length_tolerance(
+        self, gs_file: GSFile
+    ) -> Dict[str, int]:
+        """
+        Obtains the value of cropped_read_length_tolerance to post for bam files from
+        crop_length input of an arbitrary align task in the pipeline (value will be the
+        same for all align tasks since the tolerance is a global parameter). If the
+        crop_length in the pipeline is 0, then no cropping was performed and the
+        cropped_read_length_tolerance will not be posted (return empty dict).
+
+        Note that here we are assuming the crop length will always be the same for all
+        of the align tasks
+        """
+        align_task = self.analysis.get_tasks(task_name="align")[0]
+        crop_length = align_task.inputs["crop_length"]
+        crop_length_tol = align_task.inputs["crop_length_tol"]
+        if crop_length == 0:
+            return {}
+        return {"cropped_read_length_tolerance": crop_length_tol}
+
     def make_chip_alignment_qc(self, encode_file: EncodeFile, gs_file: GSFile) -> None:
         """
         This function typecasts to match the ENCODE schema. Trucated zero values could
